#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <string.h>
#include <sys/select.h>
#include <sys/types.h>

#include <PODEnv.h>

#define TRUE 1
#define FALSE 0

#define LEFT -1
#define FORWARD 0
#define RIGHT 1

#define SPEED 250
#define DISTANCE_THRESHOLD 4
#define BRIGHTNESS_THRESHOLD 50

class Robot : public PODEnv
{
   public:
      Robot();

      int Forward(unsigned char Speed);
      int Backward(unsigned char Speed);
      int Left();
      int Right();
      int Stop();

      float Heading();
      int Distance();
      int DisplayLightReadings();

      int ChooseBrightest(int &Brightness);
      int SeekBrightest();

   protected:
      MotorControlPOD *LeftMotor;
      MotorControlPOD *RightMotor;

      LightSensorPOD *LeftSensor;
      LightSensorPOD *CenterSensor;
      LightSensorPOD *RightSensor;

      SonarPOD *Sonar;
      CompassPOD *Compass;

      AlarmPOD *Alarm;
   private:
      POD *FindPOD(int POD_Type, char *POD_Name);
};

Robot::Robot() : PODEnv(POD_ENV_OPTION_TEXT_STATUS)
{

   LeftMotor = dynamic_cast <MotorControlPOD *>
      (FindPOD(POD_TYPE_MOTORCONTROL, "Left"));
   RightMotor = dynamic_cast <MotorControlPOD *>
      (FindPOD(POD_TYPE_MOTORCONTROL, "Right"));

   LeftSensor = dynamic_cast <LightSensorPOD *>
      (FindPOD(POD_TYPE_LIGHTSENSOR, "Left"));
   RightSensor = dynamic_cast <LightSensorPOD *>
      (FindPOD(POD_TYPE_LIGHTSENSOR, "Right"));
   CenterSensor = dynamic_cast <LightSensorPOD *>
      (FindPOD(POD_TYPE_LIGHTSENSOR, "Front"));

   Sonar = dynamic_cast <SonarPOD *>
      (FindPODByType(POD_TYPE_DISTANCE));
   Compass = dynamic_cast <CompassPOD *>
      (FindPODByType(POD_TYPE_COMPASS));

   Alarm = dynamic_cast <AlarmPOD *>
      (FindPODByType(POD_TYPE_ALARM));
}


POD *Robot::FindPOD(int POD_Type, char *POD_Name)
{
   POD *TempPOD;

   TempPOD = FindPODByName(POD_Type, POD_Name);
   if (! TempPOD)
   {
      printf("Could not find POD named \"%s\"\n", POD_Name);
//      printf("Could not find \"%s\" POD\r", POD_Text);
      return(TempPOD);
   }
   return(TempPOD);
}

int Robot::Forward(unsigned char Speed)
{
   RightMotor->Clockwise(Speed);
   LeftMotor->Clockwise(Speed);

   return(0);
}

int Robot::Backward(unsigned char Speed)
{
   RightMotor->CounterClockwise(Speed);
   LeftMotor->CounterClockwise(Speed);

   return(0);
}

int Robot::Left()
{
   RightMotor->CounterClockwise(SPEED);
   LeftMotor->Clockwise(SPEED);

   return(0);
}

int Robot::Right()
{
   RightMotor->Clockwise(SPEED);
   LeftMotor->CounterClockwise(SPEED);

   return(0);
}

int Robot::Stop()
{
   RightMotor->Stop();
   LeftMotor->Stop();

   return(0);
}

float Robot::Heading()
{
   return(Compass->GetHeading());
}

int Robot::Distance()
{
   return(Sonar->GetDistanceInches());
}

int Robot::ChooseBrightest(int &Brightness)
{
   int Left;
   int Right;
   int Center;

   Left = LeftSensor->GetIntensity();
   Center = CenterSensor->GetIntensity();
   Right = RightSensor->GetIntensity();

printf("L: %3d, R: %3d, C: %3d ", Left, Right, Center);
fflush(stdout);

   if ((Left > Center) && (Left > Right))
   {
      Brightness = Left;
      return(LEFT);
   }
   if ((Right > Center) && (Right > Left))
   {
      Brightness = Right;
      return(RIGHT);
   }
   Brightness = Center;
   return(FORWARD);
}

int Robot::SeekBrightest()
{
   int LastBrightest;
   int Brightest;
   int LastBrightness;
   int Brightness;
   int Distance;
   int Done = FALSE;

   while (! Done)
   {
      Brightest = ChooseBrightest(Brightness);
      if (Brightest == LEFT)
      {
         if (LastBrightest != LEFT)
         {
            Stop();
            Left();
         }
         else
         {
         }
         LastBrightest = Brightest;
         LastBrightness = Brightness;
         printf("Turning Left\r");
         fflush(stdout);
      }
      else if (Brightest == RIGHT)
      {
         if (LastBrightest != RIGHT)
         {
            Stop();
            Right();
         }
         else
         {
         }
         LastBrightest = Brightest;
         LastBrightness = Brightness;
         printf("Turning Right\r");
         fflush(stdout);
      }
      else /* forward */
      {
         if (LastBrightest != FORWARD)
         {
            Stop();
            Forward(SPEED);
         }
         else
         {
            Distance = Sonar->GetDistanceInches();
            if (Distance <= DISTANCE_THRESHOLD)
            {
               Stop();
               printf("Forward is brightest, but it's close-range.  Stopping\n");
               Done = TRUE;
            }
            if ((Brightness + BRIGHTNESS_THRESHOLD) < LastBrightness) /* getting darker again */
            {
               Stop();
               printf("Getting darker again.  Stopping\n");
               Done = TRUE;
            }
         }
         LastBrightest = Brightest;
         LastBrightness = Brightness;
         if (! Done) printf("Forward\r");
         fflush(stdout);
      }
   }

   Stop();
   printf("Forward should now be brightest\n");
   return(0);
}

int Robot::DisplayLightReadings()
{
   printf("Left: %d\n", LeftSensor->GetIntensity());
   printf("Center: %d\n", CenterSensor->GetIntensity());
   printf("Right: %d\n", RightSensor->GetIntensity());

   return(0);
}

int main(int argc, char *argv[])
{
   Robot *MyRobot;

   MyRobot = new Robot();

//   MyRobot->Forward(100);
//   MyRobot->Left();
//   usleep(100000);
//   MyRobot->Stop();

//   MyRobot->DisplayLightReadings();
   MyRobot->SeekBrightest();
   MyRobot->Stop();

   printf("-----------------------\n");
   MyRobot->DisplayLightReadings();

   return(0);
}
