#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <string.h>
#include <sys/select.h>
#include <sys/types.h>

#include <PODEnv.h>

/* our extended LightPOD */
#include "ColoredLightPOD.h"

int CaughtQuitSignal = 0;

void TermHandler(int sig)
{
//   printf("Caught TERM signal\n\r");
   CaughtQuitSignal = 1;
}
void IntHandler(int sig)
{
//   printf("Caught INT signal\n\r");
   CaughtQuitSignal = 1;
}

typedef struct
{
   ColoredLightPOD *Lights;
   AlarmPOD *Alarm;
   ThermometerPOD *Thermometer;
   ButtonsPOD *Buttons;
   LightSensorPOD *LightSensor;
   PowerSensorPOD *PowerSensor;
}SystemPODsStruct;

typedef struct
{
   float CurrentTemperature;
   float MaxTemp;
   int LightLevel;
   int OverTemperature;
   int PowerFail;
   int AlarmButton;
   int AlertAck;
   int EmailSent;
   int KeyPressed;
   char *Email;
   float Hysteresis;
}StatusStruct;

int ErrorNoPOD(char *PODType, int Quit = 1)
{
   printf("Could not find \"%s\" POD\r", PODType);
   if (Quit)
   {
      printf("\n");
      exit(-1);
   }
   return(0);
}

int DetectPODs(SystemPODsStruct &SystemPODs, int Quit = 1)
{
   PODEnv *MyPODs;
   POD *TempPOD;
   int TempDeviceNumber;

   MyPODs = new PODEnv(POD_ENV_OPTION_TEXT_STATUS);

   TempPOD = MyPODs->FindPODByType(POD_TYPE_LIGHT);
   if (! TempPOD)
   {
      ErrorNoPOD("LED Lights", Quit);
      return(-1);
   }
   TempDeviceNumber = TempPOD->GetDeviceNumber();
   SystemPODs.Lights = new ColoredLightPOD(TempDeviceNumber);
   
   SystemPODs.Alarm = dynamic_cast <AlarmPOD *>
      (MyPODs->FindPODByType(POD_TYPE_ALARM));
   if (! SystemPODs.Alarm)
   {
      ErrorNoPOD("Alarm", Quit);
      return(-1);
   }

   SystemPODs.Thermometer = dynamic_cast <ThermometerPOD *>
      (MyPODs->FindPODByType(POD_TYPE_THERMOMETER));
   if (! SystemPODs.Thermometer)
   {
      ErrorNoPOD("Thermometer", Quit);
      return(-1);
   }

   SystemPODs.Buttons = dynamic_cast <ButtonsPOD *>
      (MyPODs->FindPODByType(POD_TYPE_BUTTONS));
   if (! SystemPODs.Buttons)
   {
      ErrorNoPOD("Buttons", Quit);
      return(-1);
   }

   SystemPODs.LightSensor = dynamic_cast <LightSensorPOD *>
      (MyPODs->FindPODByType(POD_TYPE_LIGHTSENSOR));
   if (! SystemPODs.LightSensor)
   {
      ErrorNoPOD("Light Sensor", Quit);
      return(-1);
   }

   SystemPODs.PowerSensor = dynamic_cast <PowerSensorPOD *>
      (MyPODs->FindPODByType(POD_TYPE_POWERSENSOR));
   if (! SystemPODs.PowerSensor)
   {
      ErrorNoPOD("Power Sensor", Quit);
      return(-1);
   }
   return(0);
}

int SetupKeyboard(int Modified = 0)
{
   struct termios TermIO;
   static struct termios OrigTermIO;
   static int Configured = 0;

   tcgetattr(0, &TermIO);
   if (! Configured)
   {
      tcgetattr(0, &OrigTermIO);
      Configured ++; /* partially configured */
   }
   if (Modified == 0)
   {
      tcsetattr(0, TCSANOW, &OrigTermIO);
      return(0);
   }
   else
   {
      if (Configured < 2) /* not completely configured yet */
      {
         TermIO.c_oflag &= ~(OPOST | OPOST | ONLCR);
//         TermIO.c_cflag &= ~(PARENB | CSTOPB | CSIZE);
//         TermIO.c_cflag |= CS8 | CREAD | CLOCAL | BRKINT;
         TermIO.c_cflag = CS8 | CREAD | CLOCAL | BRKINT;
         TermIO.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHOCTL | ECHOKE);
//         TermIO.c_iflag &= ~(ICRNL | IXON);
         TermIO.c_iflag = IEXTEN;
         tcsetattr(0,TCSANOW,&TermIO);
         Configured ++;
      }
   }
   return(0);
}

/* Returns -1 if no input, or 0-255 if a character was waiting */
int GetKB()
{
   int rval;
   fd_set rsel;
   char KBInput;
   struct timeval Timeout;

   Timeout.tv_sec = 0;
   Timeout.tv_usec = 1;

   SetupKeyboard(1);

   FD_ZERO(&rsel);
   FD_SET(0, &rsel);
   rval = select(32, &rsel, 0, 0, &Timeout);
   if (FD_ISSET(0, &rsel) != 0)
   {
      read(0, &KBInput, 1);
      return(KBInput);
   }
   return(-1);
}

float GetTemperature(SystemPODsStruct &SystemPODs)
{
   float CurrentTemperature;

   SystemPODs.Thermometer->GetTemperature(CurrentTemperature);
   return(CurrentTemperature);
}

int PowerOn(SystemPODsStruct &SystemPODs)
{
   return(SystemPODs.PowerSensor->GetPowerStatus());
}

int AlarmOn(StatusStruct &Status)
{
   if (Status.AlertAck) return(0);
   if (! Status.AlarmButton) return(0);
   return(1);
}

int SendEmail(StatusStruct &Status)
{
   FILE *EmailProgram;
   char Command[500];

   if (! Status.Email)
   {
//      printf("\n\rNo Email Address set\n");
//      fflush(stdout);
      return(-1);
   }
   printf("\n\r");
   sprintf(Command, "/usr/sbin/sendmail -bm %s", Status.Email);
   EmailProgram = popen(Command, "w");
   if (EmailProgram)
   {
      fprintf(EmailProgram, "From: System_Monitor\nTo: %s\nSubject: Alert\n\n", Status.Email);
      fprintf(EmailProgram, "Alert...\n");
      pclose(EmailProgram);
      printf("Email Sent\n\r");
   }
   else printf("Error while sending email\n");
   fflush(stdout);
   return(0);
}

int SetEmailAddress(StatusStruct &Status)
{
   char TempString[400];
   int Counter = 0;
   int TempChar = 0;
   int Done = 0;

   if (Status.Email) free(Status.Email);
   Status.Email = NULL;

   printf("\n\rPlease enter the new email address: ");
   fflush(stdout);
   while (!Done)
   {
      TempChar = GetKB();
      if (TempChar > 0)
      {
         printf("%c", TempChar);
         fflush(stdout);
         TempString[Counter++] = TempChar;
      }
      if (Counter >= 399) Done = 1;
      else if (TempChar == '\n') Done = 1;
      else if (TempChar == '\r') Done = 1;
      TempString[Counter] = '\0';
   }
   TempString[399] = '\0';
   Status.Email = (char *)(calloc(strlen(TempString) + 1, sizeof(TempString[0])));
   strncpy(Status.Email, TempString, strlen(TempString));
   return(0);
}

int Alert(SystemPODsStruct &SystemPODs, StatusStruct &Status)
{
   if (AlarmOn(Status)) SystemPODs.Alarm->Sound(2.0);
   SendEmail(Status);
   Status.EmailSent = 1;
   return(Status.AlarmButton);
}

int IsError(StatusStruct &Status)
{
   if (Status.PowerFail | Status.OverTemperature) return(1);
   Status.AlertAck = 0;
   Status.EmailSent = 0;
   return(0);
}

int PrintStatus(StatusStruct &Status)
{
   printf("Temp: %3.2f ", Status.CurrentTemperature);
   if (Status.OverTemperature) printf("HIGH ");
   else printf("OKAY ");
   printf("(%3.2f) ", Status.MaxTemp);
   fflush(stdout);

   printf("  Power: ");
   if (Status.PowerFail) printf("FAILED ");
   else printf("OKAY   ");
   fflush(stdout);

   printf("  Light: %4d ", Status.LightLevel);
   if (Status.LightLevel > 400) printf("ON  ");
   else printf("OFF ");
   fflush(stdout);

   printf("  Alert: ");
   if (IsError(Status))
   {
      if (! Status.AlarmButton) printf("Silenced     ");
      else if (Status.AlertAck) printf("Acknowledged ");
      else printf("Sounding     ");
   }
   else printf("NONE         ");
   fflush(stdout);

   printf("Email: ");
   if (Status.Email) printf("%s", Status.Email);
   else printf("Not Set");

   printf("\r");
   fflush(stdout);
   return(0);
}

int ShowSystemStatus(SystemPODsStruct &SystemPODs, StatusStruct &Status, StatusStruct &PreviousStatus, int SystemOn = 1)
{
   if (SystemOn) SystemPODs.Lights->Green(1);
   else SystemPODs.Lights->Green(0);

   if (Status.PowerFail) SystemPODs.Lights->Red(1);
   else SystemPODs.Lights->Red(0);

   if (Status.OverTemperature) SystemPODs.Lights->Yellow(1);
   else SystemPODs.Lights->Yellow(0);

   if (IsError(Status)) Alert(SystemPODs, Status);
   PrintStatus(Status);
   return(0);
}

int main(int argc, char *argv[])
{
   SystemPODsStruct SystemPODs;
   StatusStruct Status;
   StatusStruct PreviousStatus;
   int Quit = 0;

   Status.OverTemperature = 0;
   Status.PowerFail = 0;
   Status.AlarmButton = 0;
   Status.AlertAck = 0;
   Status.KeyPressed = 0;
   Status.MaxTemp = 77.0;
   Status.Hysteresis = 0.5;
   Status.Email = NULL;
//   Status.Email = (char *)(calloc(strlen("sorenson@pel.cs.byu.edu") + 1, 1));
//   strncpy(Status.Email, "sorenson@pel.cs.byu.edu", strlen("sorenson@pel.cs.byu.edu"));

   signal(SIGTERM, TermHandler);
   signal(SIGINT, IntHandler);

   while (DetectPODs(SystemPODs, 0) < 0)
   {
      printf("\nSleeping...\n");
      fflush(stdout);
      sleep(10);
   }

   while (!Quit)
   {
      SystemPODs.Thermometer->GetTemperature(Status.CurrentTemperature);
      Status.AlarmButton = SystemPODs.Buttons->GetStatus(3);
      if (Status.OverTemperature)
      {
         if (Status.CurrentTemperature > (Status.MaxTemp - (Status.Hysteresis / 2.0))) Status.OverTemperature = 1;
         else Status.OverTemperature = 0;
      }
      else
      {
         if (Status.CurrentTemperature > (Status.MaxTemp + (Status.Hysteresis / 2.0))) Status.OverTemperature = 1;
         else Status.OverTemperature = 0;
      }

      if (PowerOn(SystemPODs)) Status.PowerFail = 0;
      else Status.PowerFail = 1;

      Status.LightLevel = SystemPODs.LightSensor->GetIntensity();
      Status.KeyPressed = GetKB();
      ShowSystemStatus(SystemPODs, Status, PreviousStatus);
      if (Status.KeyPressed == 'q' || Status.KeyPressed == 'Q') Quit = 1;
      else if (Status.KeyPressed == 't') Status.MaxTemp -= 0.25;
      else if (Status.KeyPressed == 'T') Status.MaxTemp += 0.25;
      else if (Status.KeyPressed == 'a') Status.AlertAck = 1;
      else if (Status.KeyPressed == 'e' || Status.KeyPressed == 'E') SetEmailAddress(Status);

      if (CaughtQuitSignal) Quit = 1;
   }
   SetupKeyboard(0);
   printf("\n\r");
   fflush(stdout);
}
