#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <string.h>
#include <sys/select.h>
#include <sys/types.h>

#include <PODEnv.h>

/* our extended LightPOD */
#include "ColoredLightPOD.h"

#define TRUE 1
#define FALSE !TRUE

#define ON 1
#define OFF 0

#define OK 1
#define FAIL 0

#define EMAIL_ADDRESS "sorenson@cs.byu.edu"
#define HYSTERESIS 0.25
#define MAX_TEMPERATURE 75.0

#define LIGHT_THRESHOLD 400

#define SOUND_LENGTH 2.0
#define BUTTON_ONOFF 3
#define BUTTON_PUSHBUTTON 0

typedef struct
{
   ColoredLightPOD *Lights;
   AlarmPOD *Alarm;
   ThermometerPOD *Thermometer;
   ButtonsPOD *Buttons;
   LightSensorPOD *LightSensor;
   PowerSensorPOD *PowerSensor;
}SystemPODsStruct;

typedef struct
{
   int OldLightLevel;
   int CurrentLightLevel;

   int OldPower;
   int CurrentPower;

   float OldTemperature;
   float CurrentTemperature;

   int AlertAcknowledged;

}StatusStruct;

POD *FindPOD(PODEnv *MyPODs, int POD_Type, char *POD_Text)
{
   POD *TempPOD;

   TempPOD = MyPODs->FindPODByType(POD_Type);
   if (! TempPOD)
   {
      printf("Could not find \"%s\" POD\r", POD_Text);
      return(TempPOD);
   }
   return(TempPOD);
}

int DetectPODs(SystemPODsStruct &SystemPODs)
{
   PODEnv *MyPODs;
   POD *TempPOD;
   int TempDeviceNumber;

   MyPODs = new PODEnv(POD_ENV_OPTION_TEXT_STATUS);

   TempPOD = FindPOD(MyPODs, POD_TYPE_LIGHT, "LED Lights");
   if (! TempPOD) return(FAIL);
   TempDeviceNumber = TempPOD->GetDeviceNumber();
   SystemPODs.Lights = new ColoredLightPOD(TempDeviceNumber);

   SystemPODs.Alarm = dynamic_cast <AlarmPOD *>
      (FindPOD(MyPODs, POD_TYPE_ALARM, "Alarm"));
   if (! SystemPODs.Alarm) return(FAIL);

   SystemPODs.Thermometer = dynamic_cast <ThermometerPOD *>
      (FindPOD(MyPODs, POD_TYPE_THERMOMETER, "Thermometer"));
   if (! SystemPODs.Thermometer) return(FAIL);

   SystemPODs.Buttons = dynamic_cast <ButtonsPOD *>
      (FindPOD(MyPODs, POD_TYPE_BUTTONS, "Buttons"));
   if (! SystemPODs.Buttons) return(FAIL);

   SystemPODs.LightSensor = dynamic_cast <LightSensorPOD *>
      (FindPOD(MyPODs, POD_TYPE_LIGHTSENSOR, "Light Sensor"));
   if (! SystemPODs.LightSensor) return(FAIL);
   SystemPODs.LightSensor->SetThreshold(LIGHT_THRESHOLD);

   SystemPODs.PowerSensor = dynamic_cast <PowerSensorPOD *>
      (FindPOD(MyPODs, POD_TYPE_POWERSENSOR, "Power Sensor"));
   if (! SystemPODs.PowerSensor) return(FAIL);

   return(OK);
}

int SendEmail(char *Message)
{
   FILE *EmailProgram;
   char Command[500];

   sprintf(Command, "/usr/sbin/sendmail -bm %s", EMAIL_ADDRESS);
   EmailProgram = popen(Command, "w");
   if (EmailProgram)
   {
      fprintf(EmailProgram, "From: System_Monitor\nTo: %s\nSubject: Alert\n\n", EMAIL_ADDRESS);
      fprintf(EmailProgram, "Alert: %s\n", Message);
      pclose(EmailProgram);
      printf("Email \"Alert: %s\" Sent\n\r", Message);
   }
   else printf("Error while sending email\n");
   fflush(stdout);
   return(OK);
}

int ShowSystemStatus(SystemPODsStruct &SystemPODs, StatusStruct &Status)
{
   int Error = FALSE;

   SystemPODs.Lights->Green(ON); /* system operational */

   Status.CurrentPower = SystemPODs.PowerSensor->GetPowerStatus();
   if (! Status.CurrentPower) /* power has failed */
   {
      printf("Power OFF ");
      SystemPODs.Lights->Red(ON);
      Error = TRUE;
      if (Status.OldPower != Status.CurrentPower) /* failed right now */
      {
         SendEmail("Power Failed");
         Status.AlertAcknowledged = FALSE;
      }
   }
   else SystemPODs.Lights->Red(OFF);
   Status.OldPower = Status.CurrentPower;


   SystemPODs.Thermometer->GetTemperature(Status.CurrentTemperature);
   if (Status.CurrentTemperature > (MAX_TEMPERATURE + HYSTERESIS)) /* high */
   {
      printf("Temperature HIGH ");
      SystemPODs.Lights->Yellow(ON); /* indicate high temperature */
      Error = TRUE;
      if (Status.OldTemperature <= (MAX_TEMPERATURE - HYSTERESIS)) /* new problem */
      {
         SendEmail("High Temperature");
         Status.AlertAcknowledged = FALSE;
      }
      Status.OldTemperature = Status.CurrentTemperature;
   }
   else /* temperature normal */
   {
      SystemPODs.Lights->Yellow(OFF);
      if (Status.CurrentTemperature <= (MAX_TEMPERATURE - HYSTERESIS))
         Status.OldTemperature = Status.CurrentTemperature;
   }

   Status.CurrentLightLevel = SystemPODs.LightSensor->IsLightOn();

   if (Error == TRUE) /* problem...sound the alarm */
   {                  /* unless alert has been acknowledged */
      if ((Status.OldLightLevel == FALSE)
         && (Status.CurrentLightLevel == TRUE))
      {
         Status.AlertAcknowledged = TRUE;
         printf("Acknowledged by lights being turned on\n");
      }

      if (SystemPODs.Buttons->GetStatus(BUTTON_PUSHBUTTON))
      {
         Status.AlertAcknowledged = TRUE;
         printf("Acknowledged by button\n");
      }

      if ((! Status.AlertAcknowledged)
         && SystemPODs.Buttons->GetStatus(BUTTON_ONOFF))
      {
         SystemPODs.Alarm->Sound(SOUND_LENGTH);
         printf("Alarm ON                           \r");
      }
      else printf("Alarm OFF                           \r");
   }
   else
   {
      printf("Status NORMAL                                      \r");
      Status.AlertAcknowledged = TRUE;
   }
   Status.OldLightLevel = Status.CurrentLightLevel;

   fflush(stdout);
   return(OK);
}

int main(int argc, char *argv[])
{
   SystemPODsStruct SystemPODs;
   StatusStruct Status;

   while (DetectPODs(SystemPODs) < 0)
   {
      printf("\nSleeping...\n");
      fflush(stdout);
      sleep(10);
   }

   Status.OldPower = TRUE;
   Status.OldTemperature = 0;

   while (TRUE)
   {
      ShowSystemStatus(SystemPODs, Status);
   }
}
